

def converse_handler():
    """Decorator for aliasing a method as the converse method"""

    def real_decorator(func):
        if not hasattr(func, 'converse'):
            func.converse = True
        return func

    return real_decorator



import re
import time
from typing import List

from ovos_utils.intents import IntentBuilder, Intent
from ovos_utils.log import LOG
from ovos_utils.messagebus import Message, dig_for_message
from ovos_utils.skills import get_non_properties
from ovos_utils.skills.audioservice import AudioServiceInterface
from ovos_utils.skills.settings import PrivateSettings
from ovos_utils.sound import play_audio

from ovos_workshop.decorators.killable import killable_event, \
    AbortQuestion
from ovos_workshop.skills.layers import IntentLayers
from ovos_workshop.skills.mycroft_skill import MycroftSkill, is_classic_core


class OVOSSkill(MycroftSkill):
    """
    New features:
        - all patches for MycroftSkill class
        - self.private_settings
        - killable intents
        - IntentLayers
    """

    def __init__(self, *args, **kwargs):
        # note - define these before super() because of self.bind()
        self.private_settings = None
        self._threads = []
        self._original_converse = self.converse
        self.intent_layers = IntentLayers()
        self.audio_service = None
        super(OVOSSkill, self).__init__(*args, **kwargs)

    def bind(self, bus):
        super().bind(bus)
        if bus:
            # here to ensure self.skill_id is populated
            self.private_settings = PrivateSettings(self.skill_id)
            self.intent_layers.bind(self)
            self.audio_service = AudioServiceInterface(self.bus)

    # new public api, these are not available in MycroftSkill
    @property
    def is_fully_initialized(self):
        """Determines if the skill has been fully loaded and setup.
        When True all data has been loaded and all internal state and events setup"""
        return self._is_fully_initialized

    @property
    def stop_is_implemented(self):
        return self._stop_is_implemented

    @property
    def converse_is_implemented(self):
        return self._converse_is_implemented

    def activate(self):
        """Bump skill to active_skill list in intent_service.
        This enables converse method to be called even without skill being
        used in last 5 minutes.
        """
        self._activate()

    def deactivate(self):
        """remove skill from active_skill list in intent_service.
        This stops converse method from being called
        """
        self._deactivate()

    def play_audio(self, filename):
        if not is_classic_core():
            try:
                from mycroft.version import OVOS_VERSION_BUILD, OVOS_VERSION_MINOR, OVOS_VERSION_MAJOR
                if OVOS_VERSION_MAJOR >= 1 or \
                        OVOS_VERSION_MINOR > 0 or \
                        OVOS_VERSION_BUILD >= 4:
                    self.bus.emit(Message("mycroft.audio.queue",
                                          {"filename": filename}))
                    return
            except:
                pass
        LOG.warning("self.play_audio requires ovos-core >= 0.0.4a45, falling back to local skill playback")
        play_audio(filename).wait()

    @property
    def core_lang(self):
        """Get the configured default language."""
        return self._core_lang

    @property
    def secondary_langs(self):
        """Get the configured secondary languages, mycroft is not
        considered to be in these languages but i will load it's resource
        files. This provides initial support for multilingual input"""
        return self._secondary_langs

    @property
    def native_langs(self):
        """Languages natively supported by core
        ie, resource files available and explicitly supported
        """
        return self._native_langs

    @property
    def alphanumeric_skill_id(self):
        """skill id converted to only alphanumeric characters
         Non alpha-numeric characters are converted to "_"

        Returns:
            (str) String of letters
        """
        return self._alphanumeric_skill_id

    @property
    def resources(self):
        """Instantiates a ResourceFileLocator instance when needed.
        a new instance is always created to ensure self.lang
        reflects the active language and not the default core language
        """
        return self._resources

    def load_lang(self, root_directory=None, lang=None):
        """Instantiates a ResourceFileLocator instance when needed.
        a new instance is always created to ensure lang
        reflects the active language and not the default core language
        """
        return self._load_lang(root_directory, lang)

    def voc_match(self, *args, **kwargs):
        try:
            return super().voc_match(*args, **kwargs)
        except FileNotFoundError:
            return False

    def voc_list(self, voc_filename, lang=None) -> List[str]:
        """
        Get vocabulary list and cache the results

        Args:
            voc_filename (str): Name of vocabulary file (e.g. 'yes' for
                                'res/text/en-us/yes.voc')
            lang (str): Language code, defaults to self.lang

        Returns:
            list: List of vocabulary found in voc_filename
        """
        return self._voc_list(voc_filename, lang)

    def remove_voc(self, utt, voc_filename, lang=None):
        """ removes any entry in .voc file from the utterance """
        if utt:
            # Check for matches against complete words
            for i in self.voc_list(voc_filename, lang):
                # Substitute only whole words matching the token
                utt = re.sub(r'\b' + i + r"\b", "", utt)

        return utt

    def _register_decorated(self):
        """Register all intent handlers that are decorated with an intent.

        Looks for all functions that have been marked by a decorator
        and read the intent data from them.  The intent handlers aren't the
        only decorators used.  Skip properties as calling getattr on them
        executes the code which may have unintended side-effects
        """
        super()._register_decorated()
        for attr_name in get_non_properties(self):
            method = getattr(self, attr_name)
            if hasattr(method, 'intent_layers'):
                for layer_name, intent_files in \
                        getattr(method, 'intent_layers').items():
                    self.register_intent_layer(layer_name, intent_files)

            # TODO support for multiple converse handlers (?)
            if hasattr(method, 'converse'):
                self.converse = method

    def register_intent_layer(self, layer_name, intent_list):
        for intent_file in intent_list:
            if IntentBuilder is not None and isinstance(intent_file, IntentBuilder):
                intent = intent_file.build()
                name = intent.name
            elif Intent is not None and isinstance(intent_file, Intent):
                name = intent_file.name
            else:
                name = f'{self.skill_id}:{intent_file}'
            self.intent_layers.update_layer(layer_name, [name])

    # killable_events support
    def send_stop_signal(self, stop_event=None):
        msg = dig_for_message() or Message("mycroft.stop")
        # stop event execution
        if stop_event:
            self.bus.emit(msg.forward(stop_event))

        # stop TTS
        self.bus.emit(msg.forward("mycroft.audio.speech.stop"))

        # Tell ovos-core to stop recording (not in mycroft-core)
        self.bus.emit(msg.forward('recognizer_loop:record_stop'))

        # special non-ovos handling
        if is_classic_core():
            # NOTE: mycroft does not have an event to stop recording
            # this attempts to force a stop by sending silence to end STT step
            self.bus.emit(Message('mycroft.mic.mute'))
            time.sleep(1.5)  # the silence from muting should make STT stop recording
            self.bus.emit(Message('mycroft.mic.unmute'))

        time.sleep(0.5)  # if TTS had not yet started
        self.bus.emit(msg.forward("mycroft.audio.speech.stop"))


# backwards compat alias, no functional difference
class OVOSFallbackSkill(OVOSSkill):
    def __new__(cls, *args, **kwargs):
        from ovos_workshop.skills.fallback import FallbackSkill
        return FallbackSkill(*args, **kwargs)



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Common functionality relating to the implementation of mycroft skills."""

import shutil
from abc import ABCMeta
from os.path import join, exists

from ovos_utils.log import LOG
from ovos_workshop.skills.base import BaseSkill, is_classic_core


class _SkillMetaclass(ABCMeta):
    """ To override isinstance checks we need to use a metaclass """

    def __instancecheck__(self, instance):
        if is_classic_core():
            # instance imported from vanilla mycroft
            from mycroft.skills import MycroftSkill as _CoreSkill
            if issubclass(self.__class__, _CoreSkill):
                return True

        return super().__instancecheck__(instance)


class MycroftSkill(BaseSkill, metaclass=_SkillMetaclass):
    """Base class for mycroft skills providing common behaviour and parameters
    to all Skill implementations.

    For information on how to get started with creating mycroft skills see
    https://mycroft.ai/documentation/skills/introduction-developing-skills/

    New methods added here are always private, public apis for Skill class are added in OVOSSkill
    This is done to ensure no syntax errors when a MycroftSkill object comes from mycroft-core

    Args:
        name (str): skill name
        bus (MycroftWebsocketClient): Optional bus connection
        use_settings (bool): Set to false to not use skill settings at all (DEPRECATED)
    """

    def __init__(self, name=None, bus=None, use_settings=True, *args, **kwargs):
        super().__init__(name=name, bus=bus, *args, **kwargs)

        self._initial_settings = {}
        self.settings_write_path = None
        self.settings_manager = None

        # old kludge from fallback skills, unused according to grep
        if use_settings is False:
            LOG.warning("use_settings has been deprecated! skill settings are always enabled")

        if is_classic_core():
            self.settings_write_path = self.root_dir

    def _init_settings_manager(self):
        super()._init_settings_manager()
        # backwards compat - self.settings_meta has been deprecated in favor of settings manager
        if is_classic_core():
            from mycroft.skills.settings import SettingsMetaUploader
        else:
            try:  # ovos-core compat layer
                from mycroft.deprecated.skills.settings import SettingsMetaUploader
                self._settings_meta = SettingsMetaUploader(self.root_dir, self.skill_id)
            except ImportError:
                pass  # standalone skill, skip backwards compat property

    def _init_settings(self):
        """Setup skill settings."""
        if is_classic_core():
            # migrate settings if needed
            if not exists(self._settings_path) and exists(self._old_settings_path):
                LOG.warning("Found skill settings at pre-xdg location, migrating!")
                shutil.copy(self._old_settings_path, self._settings_path)
                LOG.info(f"{self._old_settings_path} moved to {self._settings_path}")

        super()._init_settings()

    # renamed in base class for naming consistency
    def init_dialog(self, root_directory=None):
        """ DEPRECATED: use load_dialog_files instead """
        self.load_dialog_files(root_directory)

    # renamed in base class for naming consistency
    def make_active(self):
        """Bump skill to active_skill list in intent_service.

        This enables converse method to be called even without skill being
        used in last 5 minutes.

        deprecated: use self.activate() instead
        """
        self._activate()

    # patched due to functional (internal) differences under mycroft-core
    def _on_event_end(self, message, handler_info, skill_data):
        """Store settings and indicate that the skill handler has completed
        """
        if not is_classic_core():
            return super()._on_event_end(message, handler_info, skill_data)

        # mycroft-core style settings
        if self.settings != self._initial_settings:
            try:
                from mycroft.skills.settings import save_settings
                save_settings(self.settings_write_path, self.settings)
                self._initial_settings = dict(self.settings)
            except Exception as e:
                LOG.exception("Failed to save skill settings")
        if handler_info:
            msg_type = handler_info + '.complete'
            message.context["skill_id"] = self.skill_id
            self.bus.emit(message.forward(msg_type, skill_data))

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate(self, text, data=None):
        """Deprecated method for translating a dialog file.
        use self._resources.render_dialog(text, data) instead"""
        return self._resources.render_dialog(text, data)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_namedvalues(self, name, delim=','):
        """Deprecated method for translating a name/value file.
        use elf._resources.load_named_value_filetext, data) instead"""
        return self._resources.load_named_value_file(name, delim)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_list(self, list_name, data=None):
        """Deprecated method for translating a list.
        use delf._resources.load_list_file(text, data) instead"""
        return self._resources.load_list_file(list_name, data)

    # renamed in base class for naming consistency
    # refactored to use new resource utils
    def translate_template(self, template_name, data=None):
        """Deprecated method for translating a template file
        use delf._resources.template_file(text, data) instead"""
        return self._resources.load_template_file(template_name, data)

    # refactored - backwards compat + log warnings
    @property
    def settings_meta(self):
        LOG.warning("self.settings_meta has been deprecated! please use self.settings_manager instead")
        return self._settings_meta

    # refactored - backwards compat + log warnings
    @settings_meta.setter
    def settings_meta(self, val):
        LOG.warning("self.settings_meta has been deprecated! please use self.settings_manager instead")
        self._settings_meta = val

    # internal - deprecated under ovos-core
    @property
    def _old_settings_path(self):
        old_dir = self.config_core.get("data_dir") or "/opt/mycroft"
        old_folder = self.config_core.get("skills", {}).get("msm", {}) \
                         .get("directory") or "skills"
        return join(old_dir, old_folder, self.skill_id, 'settings.json')

    # patched due to functional (internal) differences under mycroft-core
    @property
    def _settings_path(self):
        if is_classic_core():
            if self.settings_write_path and self.settings_write_path != self.root_dir:
                LOG.warning("self.settings_write_path has been deprecated! "
                            "Support will be dropped in a future release")
                return join(self.settings_write_path, 'settings.json')
        return super()._settings_path
